from types import UnicodeType, StringType
from DateTime import DateTime

from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from OFS.PropertyManager import PropertyManager

from Products.CMFCore.utils import getToolByName

try:
   from Products.CMFCore.permissions import View, ModifyPortalContent, setDefaultRoles
except:
   from Products.CMFCore.CMFCorePermissions import View, ModifyPortalContent, setDefaultRoles

from Products.eCampaigning.config import AccessEcampaigningDataPermission

__all__ = ["ActionRecord", "addRecord", "addRecordForm"]

class ActionRecord(SimpleItem, PropertyManager):
   "public action content"

   meta_type = "Action Record"
   content_icon   = "action.png"

   manage_options = PropertyManager.manage_options

   security = ClassSecurityInfo()

   __allow_access_to_unprotected_attributes__ = 0
   
   __ac_permissions = (

      ('Manage properties', 
        ('manage_addProperty',
         'manage_editProperties',
         'manage_delProperties',
         'manage_changeProperties'
        ),
        ("Manager",),
      ),

      ('Access contents information',
         ('hasProperty', 'propertyIds', 'propertyValues','propertyItems',
          'getProperty', 'getPropertyType', 'propertyMap', 'propertyLabel',
          'propdict', 'valid_property_id', ''
         ),
         ('Manager',),
      )
   )


   setDefaultRoles('View', ('Manager', 'Owner'))
   setDefaultRoles('Access contents information',('Manager','Owner'))


   def getFullName(self):
      "get givenname + family name"
      return getattr(self, "givenname", "") + " " +  getattr(self, "familyname", "")

   def getMessage(self, mimetype=None):
      "get message body from self or the parent action edition"      
      try:
         message = getattr(self,'message')
         # if a different mime type is asked for, get the original type & transform the message to the asked type
         if mimetype:
            portal_transforms = getToolByName(self, 'portal_transforms')
            originaltype = self.getMessagetemplate().getContentType()
            data = portal_transforms.convertTo(mimetype, message, mimetype=originaltype)
            message = data.getData()
      except AttributeError:
         accessor = self.getMessagetemplate().getField('text').accessor
         message = getattr(self.getMessagetemplate(), accessor)(mimetype=mimetype)
      

      return message


   def getDate(self):
      "return date & time"
      return DateTime(self.getProperty("datetime"))


   def getSubject(self):
      "get message subject from self or the parent action edition"
      return getattr(self, "messagesubject", self.getMessagetemplate().getMessagetitle())


   def dataToString(self, data):
      "represent objects as strings"
      if data.__class__.__name__=="DateTime":
         value = data.strftime("%y%m%d %H%M")
      elif type(data) in (list, tuple):
         value = "\n".join(data)
      else:
         value = str(data)      
      return value


   def dataToUnicode(self, data, decoders):
      "objects to unicode"

      if decoders:
         converter = self.decodeToUnicode
      else:
         converter = lambda data, decoders: unicode(data)

      if data.__class__.__name__=="DateTime":
         value = data.strftime(u'%y%m%d %H%M')
      elif type(data) in (list, tuple):
         value = converter('\n'.join(data), decoders)
      else:
         value = converter(data, decoders)
      return value


   def decodeToUnicode(self, value, decoders):
      "try decoding the string using all decoders"
      for d in decoders:
         try:
            return v.decode(d)
         except:
            pass
      raise Exception("could not decode to unicode!")


   def replaceOrEscape(self, value, textmarker, replacechar, escchar):
      "do stuff for csv output"
      if not escchar:
         # replace textmarkers
         value = value.replace(textmarker, replacechar)
      else:
         #escape the textmarkers
         value = value.replace(textmarker, escchar+textmarker)
      return textmarker + value + textmarker


   security.declareProtected(AccessEcampaigningDataPermission,"getFieldsAsText")
   def getFieldsAsText(self, fnames, separator=',', textmarker='"', replacechar="'", escchar=None, decoders=None):
      "return raw csv-formatted text, performing replacement or escaping of text marker found inside content"
      
      # First, get values from record
      values = [getattr(self, fn, '') for fn in fnames]

      # make everything a string for exporting
      values = [self.dataToString(v) for v in values]

      # we assume content has been stored using site encoding, thus no need to mess with encodings.
      # But if need be or above does not work, we could uncomment this to try different encodings.
      #values = [self.dataToUnicode(v, decoders) for v in values]

      data = []
      for value in values:

         if value is None:
            data.append(textmarker*2)
            continue
            
         # replace textmarkers in value with textmarker+escapechar if given, else replacechar
         data.append(self.replaceOrEscape(value, textmarker, replacechar, escchar))
      
      return separator.join(data)


   def _getEntryURLParameters(self):
      "parse entry URL into dict"
      try:
         query = self.entry_url().split('?')[1]
         pairs = query.split('&')
         return dict( [pair.split('=') for pair in query.split('&') if len(pair.split('='))==2] )
         
      except IndexError:
         return ()

   
   def manage_afterAdd(self, item, container):
      "do smthing?"
      SimpleItem.manage_afterAdd(self, item, container)


   def __getattr__(self, a):
      if self.hasProperty(a):
         return self.getProperty(a)
      else:
         return super.__getattr__(self, a)


InitializeClass(ActionRecord)



def addRecord(dispatcher, id):
   "constructor"
   record = ActionRecord(id)
   record._setId(id)
   dispatcher.Destination()._setObject(id, record)


def addRecordForm():
   "form constructor"
